#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>

#include <bfd.h>
#include "bdm.h"
#include "bdmlib.h"

#if 0
  #define TPU_WR16(adr,val) (*(volatile u_int16_t*)(adr)=(val))
  #define TPU_RD16(adr) (*(volatile u_int16_t*)(adr))
#else
  #define TPU_WR16(adr,val) \
  		({ \
  		  if(bdmlib_write_var(adr,BDM_SIZE_WORD,val)<0) \
  		  	goto mem_op_error; \
  		  val; \
  		})
  #define TPU_RD16(adr) \
  		({ u_int16_t temp_val; \
  		  if(bdmlib_read_var(adr,BDM_SIZE_WORD,&temp_val)<0) \
  		  	goto mem_op_error; \
  		  temp_val; \
  		})
#endif

char hashmark=0;			/* flag set by "set hash" */
int  bdm_autoreset=1;			/* automatic reset before load */
int  bdm_ttcu=0;			/* time to come up for init by rom */
char initname[255];			/* need reimplement !!!!!!!!! */
char bdm_dev_name[255]="/dev/bdm";	/* device name */

/* Numberring of TPU register */
/* First part are aliases for main CPU accesible registers  */
#define TPUREG_TSTMSRA	0x0	/* Master Shift Register A */
#define TPUREG_TSTMSRB	0x1	/* Master Shift Register B */
#define TPUREG_TSTSC	0x2	/* Test Module Shift Count */
#define TPUREG_TSTRC	0x3	/* Test Module Repetition Count */
#define TPUREG_CREG	0x4	/* Test Submodule Control Register */
#define TPUREG_DREG	0x5	/* Distributed Register */
#define TPUREG_MCR	0x6	/* TPU Module Configuration Register   */
#define    TPU_STOP  0x8000	/*    Low-Power Stop Mode Enable */
#define    TPU_TCR1P 0x6000	/*    Timer Count Register 1 Prescaler Control */
#define    TPU_TCR2P 0x1800	/*    Timer Count Register 2 Prescaler Control */
#define    TPU_EMU   0x0400	/*    Emulation Control */
#define    TPU_T2CG  0x0200	/*    TCR2 Clock/Gate Control */
#define    TPU_STF   0x0100	/*    Stop Flag */
#define    TPU_SUPV  0x0080	/*    Supervisor/Unrestricted */
#define    TPU_PSCK  0x0040	/*    Prescaler Clock */
#define    TPU_IARB  0x000f	/*    Interrupt Arbitration ID */
#define TPUREG_TCR	0x7	/* Test Configuration Register */
#define TPUREG_DSCR	0x8	/* Development Support Control Register */
#define    TPU_HOT4  0x8000	/*    Hang on T4 */
#define    TPU_BLC   0x0400	/*    Branch Latch Control */
#define    TPU_CLKS  0x0200	/*    Stop Clocks (to TCRs) */
#define    TPU_FRZ   0x0180	/*    FREEZE Assertion Response */
#define    TPU_CCL   0x0040	/*    Channel Conditions Latch */
#define    TPU_BMSK  0x003F	/*    Break cond mask */
#define    TPU_BP    0x0020	/*    Break  mPC == mPC breakpoint register */
#define    TPU_BC    0x0010	/*    Break if CHAN start or set */
#define    TPU_BH    0x0008	/*    Break if host service latch set */
#define    TPU_BL    0x0004	/*    Break if link service latch set */
#define    TPU_BM    0x0002	/*    Break if MRL set at beginning*/
#define    TPU_BT    0x0001	/*    Break if TDL set at beginnibg */
#define TPUREG_DSSR	0x9	/* Development Support Status Register   */
#define    TPU_BKPT  0x0080	/*    Breakpoint Asserted Flag */
#define    TPU_PCBK  0x0040	/*    mPC Breakpoint Flag */
#define    TPU_CHBK  0x0020	/*    Channel Register Breakpoint Flag */
#define    TPU_SRBK  0x0010	/*    Service Request Breakpoint Flag */
#define    TPU_TPUF  0x0008	/*    TPU FREEZE Flag */
#define TPUREG_TICR	0xA	/* TPU Interrupt Configuration Register */
#define    TPU_CIRL  0x0700	/*    Channel Interrupt Request Level */
#define    TPU_CIBV  0x00f0	/*    Bits [7:4] of Channel Interrupt Base Vector */
#define TPUREG_CIER	0xB	/* Channel Interrupt Enable Register */
#define TPUREG_CFSR0	0xC	/* Channel Function Select Register 0 */
#define TPUREG_HSQR0	0xD	/* Host Sequence Register 0 */
#define TPUREG_HSRR0	0xE	/* Host Service Request Register 0 */
#define TPUREG_CPR0	0xF	/* Channel Priority Register 0 */
#define TPUREG_CISR	0x10	/* Channel Interrupt Status Register */
#define TPUREG_LR	0x11	/* Link Register */
#define TPUREG_SGLR	0x12	/* Service Grant Latch Register */
#define TPUREG_DCNR	0x13	/* Decoded Channel Number Register */
/* Registers not accesible from main CPU */
#define TPUREG_P	0x14	/* Parameter Register (16-bit) */
#define TPUREG_A	0x15	/* Accumulator (16-bit) */
#define TPUREG_DIOB	0x16	/* Data I/O Buffer (16-bit) */
#define TPUREG_SR	0x17	/* Shift Register (16-bit) */
#define TPUREG_ERT	0x18	/* Event Temporary Register (16-b.) */
#define TPUREG_CHAN	0x19	/* Channel Number (4-bit) */
#define TPUREG_DEC	0x1A	/* Decrementator Reg (4-bit) */
#define TPUREG_TCR2	0x1B	/* Timebase Register 2 (16-bit) */
#define TPUREG_TCR1	0x1C	/* Timebase Register 1 (16-bit) */
#define TPUREG_PC	0x1D	/* Microprogram Counter (9-bit) */
#define TPUREG_IR	0x1E	/*  */
#define TPUREG_BPLA	0x1F	/*  */
#define TPUREG_SPLA	0x20	/*  */
#define TPUREG_PCBR	0x21	/*  */
#define TPUREG_CHANBR	0x22	/*  */
#define TPUREG_DEBUR	0x23	/* Undocumented register at 0xFFFF00 */

#define TPUREG_MAX	0x23

/* Descriptor of TPU register access */
typedef struct tpu_reg_des {
  int (*reg_rd)(struct tpu_reg_des *des, int indx);
  int (*reg_wr)(struct tpu_reg_des *des, int indx, int val);
  int adr;
  int adr1;
  int flags;
  int max_indx;
}tpu_reg_des_t;

int tpu_mem_rd16(struct tpu_reg_des *des, int indx);
int tpu_mem_wr16(struct tpu_reg_des *des, int indx, int val);
int tpu_t1_rd(struct tpu_reg_des *des, int indx);
int tpu_t1_wr(struct tpu_reg_des *des, int indx, int val);
int tpu_t2_rd(struct tpu_reg_des *des, int indx);
int tpu_t2_wr(struct tpu_reg_des *des, int indx, int val);
int tpu_reg_rd(int reg, int indx);
int tpu_reg_wr(int reg, int indx, int val);
int tpu_reg_or(int reg, int indx, int val);
int tpu_reg_and(int reg, int indx, int val);

#define TPUREG_FL_SH4	1	/* shift value by 4 bits */
#define TPUREG_FL_PMOD	2	/* modification for reg P */
#define TPUREG_FL_DIOB	4	/* modification for reg DIOB */

/* Descriptors of TPU register access */
tpu_reg_des_t tpu_regs[]={
  /* CPU accessible */
  [TPUREG_TSTMSRA]= {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFA30,},
  [TPUREG_TSTMSRB]= {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFA32,},
  [TPUREG_TSTSC]  = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFA34,},
  [TPUREG_TSTRC]  = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFA36,},
  [TPUREG_CREG]   = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFA38,},
  [TPUREG_DREG]   = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFA3A,},
  [TPUREG_MCR]	  = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFE00,},
  [TPUREG_TCR]    = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFE02,},
  [TPUREG_DSCR]   = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFE04,},
  [TPUREG_DSSR]   = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFE06,},
  [TPUREG_TICR]   = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFE08,},
  [TPUREG_CIER]   = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFE0A,},
  [TPUREG_CFSR0]  = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFE0C,},
  [TPUREG_HSQR0]  = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFE14,},
  [TPUREG_HSRR0]  = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFE18,},
  [TPUREG_CPR0]   = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFE1C,},
  [TPUREG_CISR]   = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFE20,},
  [TPUREG_LR]     = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFE22,},
  [TPUREG_SGLR]   = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFE24,},
  [TPUREG_DCNR]   = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFE26,},
  /* Special TPU - group 1 */
  [TPUREG_P]      = {reg_rd:tpu_t1_rd,reg_wr:tpu_t1_wr,adr:0x3fff,adr1:0x36ff,
  			flags:TPUREG_FL_PMOD,},
  [TPUREG_A]      = {reg_rd:tpu_t1_rd,reg_wr:tpu_t1_wr,adr:0x327f,adr1:0x361f},
  [TPUREG_DIOB]   = {reg_rd:tpu_t1_rd,reg_wr:tpu_t1_wr,adr:0,adr1:0,
  			flags:TPUREG_FL_DIOB,},
  [TPUREG_SR]     = {reg_rd:tpu_t1_rd,reg_wr:tpu_t1_wr,adr:0x347f,adr1:0x363f},
  [TPUREG_ERT]    = {reg_rd:tpu_t1_rd,reg_wr:tpu_t1_wr,adr:0x3c7f,adr1:0x365f},
  [TPUREG_CHAN]   = {reg_rd:tpu_t1_rd,reg_wr:tpu_t1_wr,adr:0x267f,adr1:0x373f,
  			flags:TPUREG_FL_SH4,},
  [TPUREG_DEC]    = {reg_rd:tpu_t1_rd,reg_wr:tpu_t1_wr,adr:0x247f,adr1:0x375f},
  [TPUREG_TCR2]   = {reg_rd:tpu_t1_rd,reg_wr:tpu_t1_wr,adr:0x387f,adr1:0x379f},
  [TPUREG_TCR1]   = {reg_rd:tpu_t1_rd,reg_wr:tpu_t1_wr,adr:0x3a7f,adr1:0x37bf},
  /* Special TPU - group 2 */
  [TPUREG_PC]     = {reg_rd:tpu_t2_rd,reg_wr:tpu_t2_wr,adr:0x9,adr1:0x1},
  [TPUREG_IR]     = {reg_rd:NULL,reg_wr:NULL,adr:0,adr1:0},
  [TPUREG_BPLA]   = {reg_rd:tpu_t2_rd,reg_wr:tpu_t2_wr,adr:0x10,adr1:0x3},
  [TPUREG_SPLA]   = {reg_rd:tpu_t2_rd,reg_wr:tpu_t2_wr,adr:0x10,adr1:0x5},
  [TPUREG_PCBR]   = {reg_rd:tpu_t2_rd,reg_wr:tpu_t2_wr,adr:0x9,adr1:0x4},
  [TPUREG_CHANBR] = {reg_rd:tpu_t2_rd,reg_wr:tpu_t2_wr,adr:0x4,adr1:0x6},
  /* Undocumented register */
  [TPUREG_DEBUR]  = {reg_rd:tpu_mem_rd16,reg_wr:tpu_mem_wr16,adr:0xFFFF00,},
};

/**************************************************************************/
/* Helper function for test register access */

static int use_fetched_reg;
static u_int16_t fetched_dscr;
static u_int16_t fetched_creg;
static u_int16_t fetched_tcr;

static u_int16_t stored_diob;
static u_int16_t stored_pc;
static u_int16_t stored_debur2;
static u_int16_t stored_x1;
static u_int16_t stored_x2;

static
void test_shift_wait(unsigned creg)
{
  int cnt=0x8000;
  tpu_reg_wr(TPUREG_CREG,0,creg|4);
  do{
    creg=tpu_reg_rd(TPUREG_CREG,0);
    if(!(creg&0x8000)) return;
  }while(cnt--);
  fprintf(stderr,"test_shift_wait : no response from shifter\n");
}

static
void test_shift_1(int arg1 /* [bp-8] */,int arg2 /* [bp-0a] */)
{
  u_int16_t dscr; /* [bp-6] */
  u_int16_t creg; /* si */
  u_int16_t tcr;  /* di */

  if (!use_fetched_reg){
    tpu_reg_wr(TPUREG_TSTMSRB,0,0);
    tpu_reg_wr(TPUREG_TSTRC,0,0);
    tpu_reg_wr(TPUREG_DREG,0,0);
    dscr=tpu_reg_rd(TPUREG_DSCR,0);
    creg=tpu_reg_rd(TPUREG_CREG,0);
    tcr=tpu_reg_rd(TPUREG_TCR,0);
  }else{
    creg=fetched_creg;
    tcr=fetched_tcr;
    dscr=fetched_dscr;
  }

  creg=(creg|0x1000)&0xFDFF;		/*!!!!!!!!!!!*/
  tpu_reg_wr(TPUREG_CREG,0,creg);

  if(!(dscr&TPU_HOT4))  /* ???????? */
    tpu_reg_wr(TPUREG_DSCR,0,dscr|TPU_HOT4);

  tcr&=0x1801;
  tpu_reg_wr(TPUREG_TCR,0,tcr|0x204);
  tpu_reg_wr(TPUREG_TSTSC,0,0x1000);
  tpu_reg_wr(TPUREG_TSTMSRA,0,arg1);

  test_shift_wait(creg);		/* test shift */

  tpu_reg_wr(TPUREG_TCR,0,tcr|0x204);
  tpu_reg_wr(TPUREG_TSTSC,0,0x1000);
  tpu_reg_wr(TPUREG_TSTMSRA,0,arg2);

  test_shift_wait(creg);		/* test shift */

  if(!(dscr&TPU_HOT4))  /* ???????? */
    tpu_reg_wr(TPUREG_DSCR,0,dscr);
}

static
void test_shift_2(u_int16_t *ret1,u_int16_t *ret2)
{
  u_int16_t dscr; /* [bp-0a] */
  u_int16_t creg; /* si */
  u_int16_t tcr;  /* di */

  if (!use_fetched_reg){
    tpu_reg_wr(TPUREG_TSTMSRB,0,0);
    tpu_reg_wr(TPUREG_TSTRC,0,0);
    tpu_reg_wr(TPUREG_DREG,0,0);
    dscr=tpu_reg_rd(TPUREG_DSCR,0);
    creg=tpu_reg_rd(TPUREG_CREG,0);
    tcr=tpu_reg_rd(TPUREG_TCR,0);
  }else{
    creg=fetched_creg;
    tcr=fetched_tcr;
    dscr=fetched_dscr;
  }

  creg=(creg|0x1000)&0xFDFF;		/*!!!!!!!!!!!*/
  tpu_reg_wr(TPUREG_CREG,0,creg);

  if(!(dscr&TPU_HOT4))  /* ???????? */
    tpu_reg_wr(TPUREG_DSCR,0,dscr|TPU_HOT4);

  tcr&=0x1801;
  tpu_reg_wr(TPUREG_TCR,0,tcr|0x220);
  tpu_reg_wr(TPUREG_TSTSC,0,0x0010);

  test_shift_wait(creg);		/* test shift */

  *ret1=tpu_reg_rd(TPUREG_TSTMSRB,0);
  tpu_reg_wr(TPUREG_TCR,0,tcr|0x224);
  tpu_reg_wr(TPUREG_TSTSC,0,0x1010);
  tpu_reg_wr(TPUREG_TSTMSRA,0,*ret1);

  test_shift_wait(creg);		/* test shift */

  *ret2=tpu_reg_rd(TPUREG_TSTMSRB,0);
  tpu_reg_wr(TPUREG_TCR,0,tcr|0x204);
  tpu_reg_wr(TPUREG_TSTSC,0,0x1000);
  tpu_reg_wr(TPUREG_TSTMSRA,0,*ret2);

  test_shift_wait(creg);		/* test shift */

  if(!(dscr&TPU_HOT4))  /* ???????? */
    tpu_reg_wr(TPUREG_DSCR,0,dscr);
}

static
int test_shift_3(int arg1 ,int arg2)
{
  u_int16_t dscr; /* [bp-8] */
  u_int16_t creg; /* si */
  u_int16_t tcr;  /* di */
  u_int16_t ret;

  if (!use_fetched_reg){
    tpu_reg_wr(TPUREG_TSTMSRB,0,0);
    tpu_reg_wr(TPUREG_TSTRC,0,0);
    tpu_reg_wr(TPUREG_DREG,0,0);
    dscr=tpu_reg_rd(TPUREG_DSCR,0);
    creg=tpu_reg_rd(TPUREG_CREG,0);
    tcr=tpu_reg_rd(TPUREG_TCR,0);
  }else{
    creg=fetched_creg;
    tcr=fetched_tcr;
    dscr=fetched_dscr;
  }

  creg=(creg|0x1000)&0xFDFF;		/*!!!!!!!!!!!*/
  tpu_reg_wr(TPUREG_CREG,0,creg);

  if(!(dscr&TPU_HOT4))  /* ???????? */
    tpu_reg_wr(TPUREG_DSCR,0,dscr|TPU_HOT4);

  tcr&=0x1801;
  tpu_reg_wr(TPUREG_TCR,0,(arg2<<4)|tcr|0x200);
  tpu_reg_wr(TPUREG_TSTSC,0,arg1);

  test_shift_wait(creg);		/* test shift */

  ret=tpu_reg_rd(TPUREG_TSTMSRB,0)>>(0x10-arg1);
  tpu_reg_wr(TPUREG_TCR,0,(arg2<<1)|tcr|0x200);
  tpu_reg_wr(TPUREG_TSTSC,0,arg1<<8);
  tpu_reg_wr(TPUREG_TSTMSRA,0,ret);

  test_shift_wait(creg);		/* test shift */

  if(!(dscr&TPU_HOT4))  /* ???????? */
    tpu_reg_wr(TPUREG_DSCR,0,dscr);

  return ret;
}

static
int test_shift_4(int arg1 ,int arg2, int val)
{
  u_int16_t dscr; /* di */
  u_int16_t creg; /* si */
  u_int16_t tcr;  /* [bp-6] */

  if (!use_fetched_reg){
    tpu_reg_wr(TPUREG_TSTMSRB,0,0);
    tpu_reg_wr(TPUREG_TSTRC,0,0);
    tpu_reg_wr(TPUREG_DREG,0,0);
    dscr=tpu_reg_rd(TPUREG_DSCR,0);
    creg=tpu_reg_rd(TPUREG_CREG,0);
    tcr=tpu_reg_rd(TPUREG_TCR,0);
  }else{
    creg=fetched_creg;
    tcr=fetched_tcr;
    dscr=fetched_dscr;
  }

  if(!(dscr&TPU_HOT4))  /* ???????? */
    tpu_reg_wr(TPUREG_DSCR,0,dscr|TPU_HOT4);

  creg=(creg|0x1000)&0xFDFF;		/*!!!!!!!!!!!*/
  tpu_reg_wr(TPUREG_CREG,0,creg);
  tcr&=0x1801;
  tpu_reg_wr(TPUREG_TCR,0,(arg2<<1)|tcr|0x200);
  tpu_reg_wr(TPUREG_TSTSC,0,arg1<<8);
  tpu_reg_wr(TPUREG_TSTMSRA,0,val);

  test_shift_wait(creg);		/* test shift */

  if(!(dscr&TPU_HOT4))  /* ???????? */
    tpu_reg_wr(TPUREG_DSCR,0,dscr);

  return 0;
}

static
void test_pha_2(int arg1,int arg2)
{
  test_shift_1(arg1,arg2);
  tpu_reg_or(TPUREG_CREG,0,0x10);
}

static
void test_pha_1()
{
  stored_debur2=tpu_reg_rd(TPUREG_DEBUR,0);
  stored_pc=tpu_reg_rd(TPUREG_PC,0);
  test_shift_2(&stored_x1,&stored_x2);
  test_pha_2(0xfc03,0x3fff);
  stored_diob=tpu_reg_rd(TPUREG_DEBUR,0);
}

static
void test_pha_3()
{
  tpu_reg_wr(TPUREG_DEBUR,0,stored_diob);
  test_pha_2(0xfc03,0x1fff);
  tpu_reg_wr(TPUREG_DEBUR,0,stored_debur2);
  tpu_reg_wr(TPUREG_PC,0,stored_pc);
  test_shift_1(stored_x1,stored_x2);
}

/**************************************************************************/
/* Other supporting routines */

static u_int16_t test_dssrdscr_1;	/* 0 => stop, 1 =>run */
static u_int16_t test_dscr_arm;
static u_int16_t test_dssr_bkpt;

int test_init_1()
{
  u_int16_t creg; /* bp-2 */
  int cnt=0x1000;

  tpu_reg_wr(TPUREG_CREG,0,0x201);
  creg=tpu_reg_rd(TPUREG_CREG,0);
  do{
    creg=tpu_reg_rd(TPUREG_CREG,0);
  }while(!(creg&1)&&(--cnt));
  tpu_reg_wr(TPUREG_CIER,0,0);
  if(!cnt){
    fprintf(stderr,"test_init_1 : busy bit always set\n");
    return -1;
  }
  return 0;
}

int test_init_2()
{
  u_int16_t dssr; /* si */
  u_int16_t dscr; /* bx */
  dssr=tpu_reg_rd(TPUREG_DSSR,0);
  dscr=tpu_reg_rd(TPUREG_DSCR,0);
  if((dssr&TPU_BKPT)||(dscr&TPU_HOT4))
    test_dssrdscr_1=0;
  else
    test_dssrdscr_1=1;
  test_dscr_arm=dscr&TPU_BMSK;
  if (dssr&TPU_BKPT)
    test_dssr_bkpt=1;
  else
    test_dssr_bkpt=0;
  return 0;
 }

/**************************************************************************/
/* Register read and write routines */

/* read TPU register - method 1 */
int tpu_t1_rd(struct tpu_reg_des *des,  int indx)
{
  unsigned val;
  test_pha_1();
  if(des->flags&TPUREG_FL_DIOB){
    val=stored_diob;			/* DIOB read */
  }else{
    if(des->flags&TPUREG_FL_PMOD)
      test_pha_2(0xf403,des->adr);	/* P read */
    else
      test_pha_2(0xfc03,des->adr);	/* rest */
    /* read value from mem 0xFFFF00 */
    val=tpu_reg_rd(TPUREG_DEBUR,0);
  }
  test_pha_3();
  if(des->flags&TPUREG_FL_SH4)
    val>>=4;
  return val;
}

/* write TPU register - method 1 */
int tpu_t1_wr(struct tpu_reg_des *des,  int indx, int val)
{
  int ret;
  test_pha_1();
  if(des->flags&TPUREG_FL_SH4)
    val<<=4;
  ret=tpu_reg_wr(TPUREG_DEBUR,0,val);
  test_pha_2(0xfc03,0x1fff);
  if(des->flags&TPUREG_FL_DIOB){
    stored_diob=val;			/* DIOB write */
  }else{
    test_pha_2(0xffff,des->adr1);	/* rest */
  }
  test_pha_3();
  return 0;
}

/* read TPU register - method 2 */
int tpu_t2_rd(struct tpu_reg_des *des,  int indx)
{
  return test_shift_3(des->adr,des->adr1);
}

/* write TPU register - method 2 */
int tpu_t2_wr(struct tpu_reg_des *des,  int indx, int val)
{
  return test_shift_4(des->adr,des->adr1,val);
}

/* read 16 bit variable from main CPU address space */
int tpu_mem_rd16(struct tpu_reg_des *des,  int indx)
{
  u_int16_t val;
  val=TPU_RD16((caddr_t)des->adr);
  return val;

  mem_op_error:
    fprintf(stderr,"tpu_mem_rd16 : read from address %06x error\n",des->adr);
    fflush(NULL);
    return -1;
}

/* write 16 bit variable to main CPU address space */
int tpu_mem_wr16(struct tpu_reg_des *des,  int indx, int val)
{
  TPU_WR16((caddr_t)des->adr,val);
  return 0;

  mem_op_error:
    fprintf(stderr,"tpu_mem_wr16 : write to address %06x error\n",des->adr);
    fflush(NULL);
    return -1;
}

/* Read TPU register */
int tpu_reg_rd(int reg, int indx)
{
  if((reg>TPUREG_MAX)||(tpu_regs[reg].reg_rd==NULL))
    return -1;
  return tpu_regs[reg].reg_rd(&tpu_regs[reg],indx);
}

/* Write TPU register */
int tpu_reg_wr(int reg, int indx, int val)
{
  if((reg>TPUREG_MAX)||(tpu_regs[reg].reg_wr==NULL))
    return -1;
  return tpu_regs[reg].reg_wr(&tpu_regs[reg],indx,val);
}

int tpu_reg_or(int reg, int indx, int val)
{
  int reg_val;
  reg_val=tpu_reg_rd(reg,indx);
  reg_val|=val;
  tpu_reg_wr(reg,indx,reg_val);
  return reg_val;
}

int tpu_reg_and(int reg, int indx, int val)
{
  int reg_val;
  reg_val=tpu_reg_rd(reg,indx);
  reg_val&=val;
  tpu_reg_wr(reg,indx,reg_val);
  return reg_val;
}

/**************************************************************************/

#define swap_l(x) (x>>24) | ((x>>8)&0xff00) | ((x<<8)&0xff0000) | ((x&0xff)<<24)

int cpu_stat()
{
  int ret;
  u_int rpc;

  ret=bdmlib_getstatus();
  printf("MCU ");
  if(ret&BDM_TARGETRESET)   printf("Reset ");
  if(ret&BDM_TARGETSTOPPED) printf("Stopped ");
  if(ret&BDM_TARGETPOWER)   printf("NoPower ");
  if(ret&BDM_TARGETNC)      printf("NoConnect ");
  printf("\n");
  if(ret&BDM_TARGETSTOPPED) {
    if((ret=bdmlib_get_sys_reg(BDM_REG_RPC, &rpc))<0) return ret;
    printf("RPC=0x%06x\n",swap_l(rpc));
  }
  return 0;
}

int main(int argc, char *argv[])
{
  int ret;

  bdmlib_setdebug(1);
  if((ret=bdmlib_open(bdm_dev_name))<0)
    { printf("bdmlib_open : %s\n",bdmlib_geterror_str(ret)); return 1; };
  bdmlib_setioctl(BDM_SPEED,0);
  if (!(bdmlib_getstatus() & BDM_TARGETSTOPPED))
    ret=bdmlib_ioctl(BDM_STOP_CHIP);
  bdmlib_set_sys_reg(BDM_REG_DFC, 5);
  bdmlib_set_sys_reg(BDM_REG_SFC, 5);

  if(cpu_stat()<0){
    printf("trying to reset !!!!!!!\n");
    ret=bdmlib_reset();
    cpu_stat();
  }

  ret=test_init_1();
  printf("test_init_1 returned %d\n",ret);
  ret=test_init_2();
  printf("test_init_2 returned %d\n",ret);
  printf("test_dssrdscr_1(run)=%d test_dscr_arm=0x%02x test_dssr_bkpt=%d\n",
         test_dssrdscr_1,test_dscr_arm,test_dssr_bkpt);

  fflush(NULL);

  printf("DCNR %04X\n",tpu_reg_rd(TPUREG_DCNR,0));
  printf("DSCR %04X\n",tpu_reg_rd(TPUREG_DSCR,0));

//  tpu_reg_or(TPUREG_CREG,0,0x1000);	/* stop microengine */
//  tpu_reg_or(TPUREG_DSCR,0,0x8200);	/* ???????????? */
  
  printf("P    %04X DIOB %04X A    %04X SR   %04X DEC  %01X\n"
  	 "ERT  %04X TCR1 %04X TCR2 %04X PC   %03X  CHAN %01X\n",
          tpu_reg_rd(TPUREG_P,0),tpu_reg_rd(TPUREG_DIOB,0),
          tpu_reg_rd(TPUREG_A,0),tpu_reg_rd(TPUREG_SR,0),
          tpu_reg_rd(TPUREG_DEC,0),tpu_reg_rd(TPUREG_ERT,0),
          tpu_reg_rd(TPUREG_TCR1,0),tpu_reg_rd(TPUREG_TCR2,0),
          tpu_reg_rd(TPUREG_PC,0),tpu_reg_rd(TPUREG_CHAN,0));

  return 0;
}